"""

This package has the following modules:
    
``build``
    This module contains functions for building the metadata file.

``miagis_schema``
    This module contains the schema for validating user input.
    
``validate``
    This module contains functions to validate the metadata file.
    
``user_input_checking``
    This module contains functions to validate user input.
    
``print_map_layers``
    This module contains the code to run the print_map_layers command.
    

"""
__version__ = "1.0.1"